# -*- coding: utf-8 -*-
"""
==================================
Color Quantization using K-Means
==================================

Performs a pixel-wise Vector Quantization (VQ) of an image of the summer palace
(China), reducing the number of colors required to show the image from 96,615
unique colors to 64, while preserving the overall appearance quality.

In this example, pixels are represented in a 3D-space and K-means is used to
find 64 color clusters. In the image processing literature, the codebook
obtained from K-means (the cluster centers) is called the color palette. Using
a single byte, up to 256 colors can be addressed, whereas an RGB encoding
requires 3 bytes per pixel. The GIF file format, for example, uses such a
palette.

For comparison, a quantized image using a random codebook (colors picked up
randomly) is also shown.
"""
# Authors: Robert Layton <robertlayton@gmail.com>
#          Olivier Grisel <olivier.grisel@ensta.org>
#          Mathieu Blondel <mathieu@mblondel.org>
#
# License: BSD 3 clause

print(__doc__)
import numpy as np
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
from sklearn.metrics import pairwise_distances_argmin
from sklearn.datasets import load_sample_image
from sklearn.utils import shuffle
from time import time

n_colors = 64

# Load the Summer Palace photo
china = load_sample_image("china.jpg")

# Convert to floats instead of the default 8 bits integer coding. Dividing by
# 255 is important so that plt.imshow behaves works well on float data (need to
# be in the range [0-1])
china = np.array(china, dtype=np.float64) / 255

# Load Image and transform to a 2D numpy array.
w, h, d = original_shape = tuple(china.shape)
assert d == 3
image_array = np.reshape(china, (w * h, d))

print("Fitting model on a small sub-sample of the data")
t0 = time()
image_array_sample = shuffle(image_array, random_state=0)[:1000]
kmeans = KMeans(n_clusters=n_colors, random_state=0).fit(image_array_sample)
print("done in %0.3fs." % (time() - t0))

# Get labels for all points
print("Predicting color indices on the full image (k-means)")
t0 = time()
labels = kmeans.predict(image_array)
print("done in %0.3fs." % (time() - t0))


codebook_random = shuffle(image_array, random_state=0)[:n_colors + 1]
print("Predicting color indices on the full image (random)")
t0 = time()
labels_random = pairwise_distances_argmin(codebook_random,
                                          image_array,
                                          axis=0)
print("done in %0.3fs." % (time() - t0))


def recreate_image(codebook, labels, w, h):
    """Recreate the (compressed) image from the code book & labels"""
    d = codebook.shape[1]
    image = np.zeros((w, h, d))
    label_idx = 0
    for i in range(w):
        for j in range(h):
            image[i][j] = codebook[labels[label_idx]]
            label_idx += 1
    return image

# Display all results, alongside original image
plt.figure(1)
plt.clf()
ax = plt.axes([0, 0, 1, 1])
plt.axis('off')
plt.title('Original image (96,615 colors)')
plt.imshow(china)

plt.figure(2)
plt.clf()
ax = plt.axes([0, 0, 1, 1])
plt.axis('off')
plt.title('Quantized image (64 colors, K-Means)')
plt.imshow(recreate_image(kmeans.cluster_centers_, labels, w, h))

plt.figure(3)
plt.clf()
ax = plt.axes([0, 0, 1, 1])
plt.axis('off')
plt.title('Quantized image (64 colors, Random)')
plt.imshow(recreate_image(codebook_random, labels_random, w, h))
plt.show()

#==============================================================================
#==============================================================================

# Copyright (c) 2016-11-02, Laurenz Wiskott, http://www.ini.rub.de/PEOPLE/wiskott/
# 2019-11-18 Converted to Python 3 by Robin Schiewer.
# License: BSD 2-Clause, https://opensource.org/licenses/BSD-2-Clause, see the end

# The part above is an unchanged copy of plot_color_quantization.py from
# http://scikit-learn.org/stable/auto_examples/cluster/plot_color_quantization.html
# It has its own copyright notice.

#==============================================================================

import math as m

SHOW_BLOCKS = True # Program should stop at each figure or run through.

#==============================================================================

# EXERCISES

#------------------------------------------------------------------------------

# EXERCISE 1: Vary the number of reference vectors (or clusters as they are
# called here) and see how the result changes.

# SOLUTION
# I simply run the code from above for different numbers of colors.

for n_colors in [2,4,8,16,32,64,128,256]:
    image_array_sample = shuffle(image_array, random_state=0)[:1000]
    kmeans = KMeans(n_clusters=n_colors, random_state=0).fit(image_array_sample)
    labels = kmeans.predict(image_array)

    fig = plt.figure(figsize=(20,10))
    plt.clf()
    ax = plt.axes([0, 0, 1, 0.95])
    plt.axis('off')
    plt.title('Quantized image ('+str(n_colors)+' colors, K-Means)')
    plt.imshow(recreate_image(kmeans.cluster_centers_, labels, w, h))
    plt.show(block=SHOW_BLOCKS)

#------------------------------------------------------------------------------

# Perform k-means clustering once again to have a defined number of reference
# vectors independly of the previous exercise.
num_samples = 1000
n_colors = 64
image_array_sample = shuffle(image_array, random_state=0)[:num_samples]
kmeans = KMeans(n_clusters=n_colors, random_state=0).fit(image_array_sample)
labels = kmeans.predict(image_array)

#------------------------------------------------------------------------------

# EXERCISE 2: Visualize how the pixel color values and reference vectors lie in
# color space.

# SOLUTION
# Since color space is three-dimensional, we can in principle perceive it as it
# is.  We just have to convey depth-information.  One way to do that is rotate
# it in depth.  This is what I do here, I animate the pixel data and reference
# vectors rotating.
speed = 1 # Modify the speed of rotation, if you like.
fig = plt.figure(figsize=(12,12))
fig.suptitle("Animation of pixel color values and reference vectors in color space.", size=16)
plt.show(block=False)
ax = fig.add_subplot(111)
for t in range(200):
    # Define angles
    phi_1 = 0.017*speed*t
    phi_2 = 0.005*speed*t
    phi_3 = 0.*speed*t # I have set this to zero because in-plane rotation does
                       # not induce depth perception.
    # Define rotation matrices (they do not commute)
    rot_1 = np.array([[1,            0,             0],
                      [0, m.cos(phi_1), -m.sin(phi_1)],
                      [0, m.sin(phi_1),  m.cos(phi_1)]])
    rot_2 = np.array([[m.cos(phi_2), 0, -m.sin(phi_2)],
                      [0,            1,             0],
                      [m.sin(phi_2), 0,  m.cos(phi_2)]])
    rot_3 = np.array([[m.cos(phi_3), -m.sin(phi_3), 0],
                      [m.sin(phi_3),  m.cos(phi_3), 0],
                      [           0,             0, 1]])
    # Combine rotation matrices
    rot = np.einsum('ij,jk->ik', rot_2, rot_1)
    rot = np.einsum('ij,jk->ik', rot_3, rot)
    ax.clear()
    # Calculate rotated projected data
    image_array_sample_rotated = np.einsum('ij,kj->ki', rot[0:2,:], image_array_sample      - [0.5,0.5,0.5]) 
    cluster_centers_rotated    = np.einsum('ij,kj->ki', rot[0:2,:], kmeans.cluster_centers_ - [0.5,0.5,0.5]) 
    # Plot data
    ax.set(xlim=[-1,1],
           ylim=[-1,1],
           xlabel='first rotated component',
           ylabel='second rotated component')
    ax.scatter(image_array_sample_rotated[:,0],
               image_array_sample_rotated[:,1],
               s=1,  marker='.', color='k')
    ax.scatter(cluster_centers_rotated[:,0],
               cluster_centers_rotated[:,1],
               s=40, marker='.', color='r')
    fig.canvas.draw()
plt.show(block=SHOW_BLOCKS)

#------------------------------------------------------------------------------

# EXERCISE 3: Do you think there are particular relationships between certain
# statistical features of the individual reference vectors?  Such features
# could be density of reference vectors, density of data points, number of data
# points within a Voronoi cell, error of the data points within a Voronoi cell,
# ...  If yes, what relationships do you expect?

# SOLUTION

# Firstly, I expect a strong positive correlation between density of data
# points and density of reference vectors.  It just pays off to place more
# reference vectors in high density regions, because one can reduce the error
# per data point more in those regions.

# Secondly, as a consequence of the higher density of reference vectors in high
# density regions I expect a smaller error per data point in those regions.

# It is not obvious to me, how the number of data points per reference vector
# should depend on the density (of data points or reference vectors).


#------------------------------------------------------------------------------

# EXERCISE 4: Test your predictions by making scatter plots between pairs of
# statistical features between which you expect a particular relationship.
# Relate your results to what you expect theoretically.

# SOLUTION

# I consider here the following statistical features for all reference vectors:
#   a) number of data points
#   b) error
#   c) error per pixel, i.e. error per data point
#   d) distance to nearest neighboring reference vector
#   e) radius within which the nearest 2 or 6 reference vectors lie
#      (neighborhood 2 and neighborhood 6, for brevity)
#   f) radius within which the nearest 10 or 100 data points lie (radius 10 and
#      radius 100, for brevity)
# I use distances rather than densities here, because they have a more well
# defined meaning and are easier to compute.

# Since I expect power-law relationships, I always plot the logarithms, and I
# plot manually determined regression lines for orientation, corresponding to
# simple quadratic, linear or square root relationships, which fit fairly well.
# Thus I assume a functional relationship between two variables 'y' and 'x' of
# the form

#           y  =     c  *     x**p
#   <=> log(y) = log(c) + log(x)*p

# with simple exponents p and some constant c.  If we know the exponent p and
# one value pair log(x) and log(y), we can calculate log(c) and c.

#       log(c) =      log(y) - log(x)*p
#   <=>     c  = exp[ log(y) - log(x)*p ]


def num_data_points_and_error_per_ref(image_array, labels, cluster_centers):
    """Function to calculate the number of data points and error per reference vector."""

    n_colors = len(cluster_centers)

    # Determine sorting order of labels.
    sort_idxs = np.argsort(labels)
    # Sort labels.
    labels_sorted = labels[sort_idxs]
    # Sort image pixels.
    image_array_sorted = image_array[sort_idxs]

    # Compute number of data points for each reference vector.
    labels_sorted_tmp = list(labels_sorted)
    num_data_points = list(range(n_colors))
    for color_idx in range(n_colors-1):
        num_data_points[color_idx] = labels_sorted_tmp.index(color_idx+1)
        labels_sorted_tmp = labels_sorted_tmp[num_data_points[color_idx]:]
    num_data_points[n_colors-1] = len(labels_sorted_tmp)

    # Compute error of data points belonging to a reference vector.
    image_array_sorted_tmp = image_array_sorted
    errors = list(range(n_colors))
    for color_idx in range(n_colors):
        errors[color_idx] = np.sum(np.einsum('ij->i',
                                                 (image_array_sorted_tmp[:num_data_points[color_idx]] 
                                                  - cluster_centers[color_idx])**2))
        image_array_sorted_tmp = image_array_sorted_tmp[num_data_points[color_idx]:]

    return num_data_points, errors


# Check plausibility of values.
num_data_points, errors = num_data_points_and_error_per_ref(image_array, labels,
                                            kmeans.cluster_centers_)
print("These two should be identical.")
print(image_array.shape[0])
print(np.sum(num_data_points))
print("These two should be similiar.")
print(kmeans.inertia_/num_samples)    # error per sample pixel
print(np.sum(errors)/np.sum(num_data_points)) # error per any image pixel 
# It is ok that the error per any image pixel is a bit higher than the error
# per sample pixel, since the reference vectors have been optimized on the
# sample pixels.  The difference is smaller for larger numbers of sample pixels
# (try num_samples = 5000 and 20000).

#..............................................................................

# Define variables.
random_state_samples = None # default None
random_state_kmeans = None # default None
num_samples = 1000 # default 1000
n_colors = 64 # default 64
trials = 6 # Number of trials.  default 6

# Define intervals of statistical features, used to plot manual regression lines.
line_radius100 = [-4.5-2, -4.5+2]
line_distance_neighbor_old = [-2.6-2*1./2., -2.6+2*1./2.]
line_distance_neighbor = [-2.8-2*2./3., -2.8+2*2./3.]
line_num_data_points = [8.4+2, 8.4-2]
line_errors_per_pixel = [-6.3-2, -6.3+2]


# FIGURE 1:

title = 'log(distance to nearest neighbor) vs log(radius 100) for all reference vectors'
fig_1 = plt.figure(figsize=(20, 10))
fig_1.suptitle(title, fontsize=30)
ax_1 = fig_1.add_subplot(111)
# Plot line with slope +1/2 for orientation. 
ax_1.plot(line_radius100, line_distance_neighbor_old, '--', color='k') 
# Plot line with slope +2/3 for orientation. 
ax_1.plot(line_radius100, line_distance_neighbor, '-', color='k') 
plt.show(block=False)

# This plot verifies that the density of reference vectors is highly correlated
# with the density of data points (both measured here in terms of distances).

# If one considers all data points, the slope in the log-log plot is close to
# 1/2 (dashed line) and the relation between distance to nearest reference
# vector and radius 100 is roughly square root.  This means, if the distance
# between the data points increases by a factor of 4, the distance between
# reference vectors increases by a factor of 2, and the number of data points
# per reference vector should go down.  By how much depends on the
# dimensionality.

# Since this exponent 1/2 leads to results that conflict with theory, I have
# also plotted a line with slope 2/3, which fits the points on the right fairly
# well and is consistent with the theory.  We will see later that the points on
# the left behave somehow strangely.  In the following, I will consider both
# alternatives.

# From the values of line_radius100 and line_distance_neighbor we can figure:

#   x = r = radius of 100 nearest data points
#   y = d = distance to nearest neighborhood
#   p = p_n = 1/2
#   c = c_n = exp[ log(y) - log(x)*p ] = exp[ (-2.6) - (-4.5)*1/2 ] 
#           = exp[-0.35] = 0.704688

# The number of data points n in a Voronoi cell decreases with the distance
# between the data points r like r**(-D), if D is the dimensionality of the
# space, and increases with the radius of the Voronoi cell (approximated here
# by the distance to the nearest neighbor) d like d**D.

#   n ~ r**(-D) * d**D

# We have seen above that d ~ r**(1/2) or r**(2/3), thus we expect

#      n ~ r**(-D) * d**D = r**(-D) * r**(D*1/2) = r**(-D*1/2) 
#   or n ~ r**(-D) * d**D = r**(-D) * r**(D*2/3) = r**(-D*1/3) 

# In 2D, i.e. for D=2, we would therefore expect an exponent of -1 or -2/3 for
# the number of data points as a function of the radius 100, and in 3D we would
# expect an exponent of -3/2 or -1.  This is tested in the next plot.


# FIGURE 2:

title = 'log(number of data points) vs log(radius 100) for all reference vectors'
fig_2 = plt.figure(figsize=(20, 10))
fig_2.suptitle(title, fontsize=30)
ax_2 = fig_2.add_subplot(111)
# Plot line with slope -1 for orientation. 
ax_2.plot(line_radius100, line_num_data_points, '-', color='k') 
plt.show(block=False)

# Plotting the number of data points vs the radius 100 confirms, as expected
# from the first figure, that the number of data points goes down as the radius
# 100 increases (and the density of data points decreases).  The slope, is -1,
# which either indicates an exponent of 1/2 in the figure above and intrinsic
# dimensionality of the data or reference vecotrs of two or and exponent of 2/3
# and the expected dimensionality 3, if we follow the arguments of the previous
# figure.  Thus, it is interesting to know whether the data points and the
# reference vectors really fill the three dimensional space or whether they lie
# on lower dimensional submanifolds.  The visualisation above already hinted at
# a lower dimensional distribution, but that was a global impression.  How does
# it look locally?


# FIGURE 3:

title = 'log(radius 10 / radius 100) vs log(radius 100) for all reference vectors'
fig_3 = plt.figure(figsize=(20, 10))
fig_3.suptitle(title, fontsize=30)
ax_3 = fig_3.add_subplot(111)
# Plot horizontal lines for orientation. 
for dim in range(1,5):
    ax_3.plot(line_radius100, [np.log(0.1**(1./dim)), np.log(0.1**(1./dim))],
              '--', color='k') 

# I use the ratio of radius 10 to radius 100 to estimate the intrinsic
# dimensionality of the data.  In 1D the radius 10 would simply be 0.1 times
# radius 100.  In 2D the factor would be the square root of 0.1, in 3D the
# cubic root of 0.1, etc.  The horizontal lines in the plot indicate these
# factors for 1D up to 4D from bottom to top.

# We find that the points for higher values of radius 100 cluster at the 3D
# line.  For low values of radius 100 the points cluster at zero, which would
# indicate infinite dimensionality, which is impossible.  This is probably due
# to a quantization of the data, so that many points have identical distance to
# the reference vector and the ratio of radius 10 and radius 100 becomes one.


# FIGURE 4:

title = 'log(neighborhood 2 / neighborhood 6) vs log(radius 100) for all reference vectors'
fig_4 = plt.figure(figsize=(20, 10))
fig_4.suptitle(title, fontsize=30)
ax_4 = fig_4.add_subplot(111)
# Plot horizontal lines for orientation. 
for dim in range(1,5):
    ax_4.plot(line_radius100, [np.log((2./6.)**(1./dim)),
                               np.log((2./6.)**(1./dim))], '--', color='k') 

# Similar plot as the previous one but for the ratio neighbor 2 to neighbor 6.
# The scatter is large, but the bulk of the data still clusters at the 3D line.

# Thus it seems the data really fills 3D, at least locally.  The plots also
# show that for lower values of radius 100, the points behave a bit strangely.
# These results suggest to favor the exponent 2/3 over 1/2 in the first plot.

# Next we consider the errors.


# FIGURE 5:

title = 'log(error per pixel) vs log(radius 100) for all reference vectors'
fig_5 = plt.figure(figsize=(20, 10))
fig_5.suptitle(title, fontsize=30)
ax_5 = fig_5.add_subplot(111)
# Plot line with slope +1 for orientation. 
ax_5.plot(line_radius100, line_errors_per_pixel, '-', color='k') 
plt.show(block=False)

# Plotting the error per data point against the radius 100 shows a linear
# dependency.  We know already that the distance to the nearest neighbor
# follows a square root in the radius 100, and since the error scales quadratic
# with the distance of the data points from the reference vector, one would
# expect a quadratic dependency on the distance to the nearest neighbor as
# well.  Taken together this yields the linear relationship between error per
# data point against the radius 100 that we see in the plot.

# Plotting the error per data point against the distance to the nearest
# neighbor would be more direct but surprisingly resulted in a slightly less
# clear correlation.

# Now, since the error per pixel goes up and the number of data points goes
# down with the radius 100, both linearly, the product of these, i.e. the
# overall error per reference vector, should be constant.  This is tested in
# the next plot.


# FIGURE 6:

title = 'log(error) vs log(radius 100) for all reference vectors'
fig_6 = plt.figure(figsize=(20, 10))
fig_6.suptitle(title, fontsize=30)
ax_6 = fig_6.add_subplot(111)

# Indeed, there is no systematic relationship between the error and the radius
# 100 of the reference vectors.


# ALL FIGURES:

plt.show(block=False)
for trial in range(trials):
    print("Trial", trial, "...")

    # Perform k-means clustering.
    image_array_sample = shuffle(image_array, random_state=random_state_samples)[:num_samples]
    kmeans = KMeans(n_clusters=n_colors, random_state=random_state_kmeans).fit(image_array_sample)
    cluster_centers = kmeans.cluster_centers_
    labels = kmeans.predict(image_array)

    # Calculate statistical features.
    num_data_points, errors = num_data_points_and_error_per_ref(image_array, labels, kmeans.cluster_centers_)
    errors_per_pixel = np.array(errors)/np.array(num_data_points)
    distance_neighbor = np.sqrt([sorted(np.einsum('ij->i',(cluster_centers - cluster_centers[cluster_idx])**2))[1] 
                    for cluster_idx in range(n_colors)])
    neighborhood2 = np.sqrt([sorted(np.einsum('ij->i',(cluster_centers - cluster_centers[cluster_idx])**2))[2]
                    for cluster_idx in range(n_colors)])
    neighborhood6 = np.sqrt([sorted(np.einsum('ij->i',(cluster_centers - cluster_centers[cluster_idx])**2))[6]
                    for cluster_idx in range(n_colors)])
    radius10 = np.sqrt([sorted(np.einsum('ij->i',(image_array - cluster_centers[cluster_idx])**2))[9]
                    for cluster_idx in range(n_colors)])
    radius100 = np.sqrt([sorted(np.einsum('ij->i',(image_array - cluster_centers[cluster_idx])**2))[99]
                    for cluster_idx in range(n_colors)])

    # Make scatter plots.
    color = ['b','r','y','k','c','m'][trial%6]
    ax_1.scatter(np.log(radius100), np.log(distance_neighbor), s=10, marker='o', color=color)
    fig_1.canvas.draw()    
    ax_2.scatter(np.log(radius100), np.log(num_data_points), s=10, marker='o', color=color)
    fig_2.canvas.draw()    
    ax_3.scatter(np.log(radius100), np.log(radius10)-np.log(radius100), s=10, marker='o', color=color)
    fig_3.canvas.draw()    
    ax_4.scatter(np.log(radius100), np.log(neighborhood2)-np.log(neighborhood6), s=10, marker='o', color=color)
    fig_4.canvas.draw()    
    ax_5.scatter(np.log(radius100), np.log(errors_per_pixel), s=10, marker='o', color=color)
    fig_5.canvas.draw()    
    ax_6.scatter(np.log(radius100), np.log(errors), s=10, marker='o', color=color)
    fig_6.canvas.draw()    

plt.show(block=SHOW_BLOCKS)

#------------------------------------------------------------------------------

# EXERCISE 5: 

# (a) There is some randomness in the quantization result.  What are the causes
# of the randomness, what is their relative contribution, and what is the
# overall randomness?

# (b) Try to quantify and visualize the amounts or randomness.

# (c) What could you do to reduce the effects of randomness.


# SOLUTION

# Ad (a): There are two causes of randomness here: (i) the choice of the sample
# pixels and (ii) the initialization of the reference vectors.  By setting the
# randomization seeds to a fixed value, as done above by setting
# 'random_state=0', one can freeze the random numbers but not get rid of the
# randomness.


# Ad (b): One can quantify the amount of (relevant) randomness by determining
# the variance in the quantization error (called 'inertia_' in 'KMeans').  And
# a simple way to visualize it is to show the resulting figure for different
# runs and see how the color values vary.

random_state_samples = None
random_state_kmeans = None
num_samples = 1000
n_colors = 64
trials = 10 # Number of trials.

trial_labels = list(range(trials))
trial_cluster_centers = list(range(trials))
trial_error = list(range(trials))
for trial in range(trials):
    print("Trial", trial, "...")
    image_array_sample = shuffle(image_array, random_state=random_state_samples)[:num_samples]
    kmeans = KMeans(n_clusters=n_colors, random_state=random_state_kmeans).fit(image_array_sample)
    trial_labels[trial] = kmeans.predict(image_array)
    trial_cluster_centers[trial] = kmeans.cluster_centers_
    num_data_points, errors = num_data_points_and_error_per_ref(image_array, trial_labels[trial], trial_cluster_centers[trial])
    trial_error[trial] = np.sum(errors)
print("Mean:", np.mean(trial_error), "; Std:",  np.std(trial_error), "(for", \
    random_state_samples, random_state_kmeans, num_samples, n_colors, trials, ")")

loops = 3 # Number of times to show all trials.
fig = plt.figure(figsize=(20,10))
fig.suptitle("Animation of the variation in the quantization.", size=16)
plt.show(block=False)
ax = fig.add_subplot(111)
for loop in range(loops):
    for trial in range(trials):
        ax.clear()
        ax.axis('off')
        ax.imshow(recreate_image(trial_cluster_centers[trial], trial_labels[trial], w, h))
        fig.canvas.draw()
plt.show(block=SHOW_BLOCKS)

# I find for the mean and standard deviation of the error the values (for
# random_state_samples, random_state_kmeans, num_samples, n_colors, trials):

# A:
#   Mean: 884.982 ; Std: 13.3302 (for None None 1000 32 20 )
#   Mean: 878.491 ; Std: 13.0616 (for None None 1000 32 20 )
#   Mean: 880.890 ; Std: 11.1770 (for None None 1000 32 20 )
#   Mean: 888.999 ; Std: 13.2197 (for None None 1000 32 20 )
#   Mean: 882.108 ; Std: 13.8905 (for None None 1000 32 20 )
#         883.094        12.9358
#           3.618         0.9227
print(np.mean([884.982,878.491,880.890,888.999,882.108]))
print( np.std([884.982,878.491,880.890,888.999,882.108]))
print(np.mean([13.3302,13.0616,11.1770,13.2197,13.8905]))
print( np.std([13.3302,13.0616,11.1770,13.2197,13.8905]))

# B:
#   Mean: 884.106 ; Std: 14.0132 (for None 0 1000 32 20 )
#   Mean: 878.448 ; Std: 12.5898 (for None 1 1000 32 20 )
#   Mean: 879.623 ; Std: 9.49085 (for None 2 1000 32 20 )
#   Mean: 887.494 ; Std: 17.6693 (for None 3 1000 32 20 )
#   Mean: 886.891 ; Std: 20.4736 (for None 4 1000 32 20 )
#         883.312        14.8473
#           3.693         3.8497
print(np.mean([884.106, 878.448, 879.623, 887.494, 886.891]))
print( np.std([884.106, 878.448, 879.623, 887.494, 886.891]))
print(np.mean([14.0132, 12.5898, 9.49085, 17.6693, 20.4736]))
print( np.std([14.0132, 12.5898, 9.49085, 17.6693, 20.4736]))

# C:
#   Mean: 866.982 ; Std: 10.6739 (for 0 None 1000 32 20 )
#   Mean: 878.798 ; Std: 11.3833 (for 1 None 1000 32 20 )
#   Mean: 873.384 ; Std: 9.58516 (for 2 None 1000 32 20 )
#   Mean: 880.833 ; Std: 10.6032 (for 3 None 1000 32 20 )
#   Mean: 885.835 ; Std: 8.32646 (for 4 None 1000 32 20 )
#         877.166        10.1144
#           6.470         1.0620
print(np.mean([866.982, 878.798, 873.384, 880.833, 885.835]))
print( np.std([866.982, 878.798, 873.384, 880.833, 885.835]))
print(np.mean([10.6739, 11.3833, 9.58516, 10.6032, 8.32646]))
print( np.std([10.6739, 11.3833, 9.58516, 10.6032, 8.32646]))

# D:
#   Mean: 872.331 ; Std: 1.1e-13 (for 0 0 1000 32 20 )
#   Mean: 875.520 ; Std: 1.1e-13 (for 1 1 1000 32 20 )
#   Mean: 867.211 ; Std: 0.0     (for 2 2 1000 32 20 )
#   Mean: 890.540 ; Std: 2.2e-13 (for 3 3 1000 32 20 )
#   Mean: 900.481 ; Std: 4.5e-13 (for 4 4 1000 32 20 )
#         881.216        1.8e-13
#          12.374        1.5e-13
print(np.mean([872.331, 875.520, 867.211, 890.540, 900.481]))
print( np.std([872.331, 875.520, 867.211, 890.540, 900.481]))
print(np.mean([1.1e-13, 1.1e-13, 0.0    , 2.2e-13, 4.5e-13]))
print( np.std([1.1e-13, 1.1e-13, 0.0    , 2.2e-13, 4.5e-13]))

# The mean values look rather similar throughout and do not seem to differ
# significantly, as one would expect, because there is no reason to assume that
# the mean should depend systematically on whether we fix the random seed or
# not.

# As far as the standard deviations are concerned, it is a bit surprising that
# they are higher in B than in A.  I do not think that this is significant,
# since they vary so much in B.  A Student-t test would help here.
# Theoretically I would not expect any difference between A and B, because if
# the samples are different each time it does not matter anymore if the seed
# for the kmeans is differen or not (not knowing the details of the
# implementation).  That the standard deviations are a bit smaller in C could
# be expected, but the difference might also not be significant, again a
# Student-t test would help.  In D the standard deviation should be zero, of
# course.


# Ad (c): The effect of randomness in the choice of samples can be reduced by
# taking more samples, which is practical within the limits of computational
# resources.


#==============================================================================

# BSD 2-Clause License

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#==============================================================================

